/*****************************************************************************
! Program to convert apeNEXT crate configuration to ILDG format
!
! author: Andre Sternbeck     <andre.sternbeck@physik.hu-berlin.de>,
!         Carsten Urbach,
!         Ines Wetzorke 
!
! usage: crate2ildg [-h] [options]  <infile> <outfile>
!          for options available see usage() below
!
!******************************************************************************/


#define _MAIN_PROGRAM

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#ifdef MPI
#include <mpi.h>
#endif
#include "geometry.h"
#include "global.h"
#include "su3.h"
#include "io.h"
#include "plaquette.h"
#include "initialize.h"


//-----------------------------------------------------------------------------
void message(char *msg)
{
  printf ( msg );
}

//-----------------------------------------------------------------------------
void die(char *msg, int exitcode)
{
  message( msg );
  exit(exitcode);
}

void usage(void)
{
    fprintf(stderr, "Usage: crate2ildg [-h] [options] infile outfile\n");
    fprintf(stderr, " options:\n"); 
    fprintf(stderr, "   -L  <L>       (default 16)\n");
    fprintf(stderr, "   -T  <T>       (default 32)\n");
    fprintf(stderr, "   -X  <N_PROCX> (default 4)\n");
    fprintf(stderr, "   -Y  <N_PROCY> (default 8)\n");
    fprintf(stderr, "   -Z  <N_PROCZ> (default 8)\n");
    fprintf(stderr, "   -b  <beta>    (default 0.0)\n");
    fprintf(stderr, "   -k  <kappa>   (default 0.0)\n");
    fprintf(stderr, "   -m  <mu>      (default 0.0)\n");
}


//-----------------------------------------------------------------------------
main ( int argc, char *arg[] ) 
{

  double plaq;
  char *infile, *outfile;
  int   status;

  g_mu = 0.0; g_beta = 0.0; g_kappa = 0.0;

  // default values of (crate) constant

  L = 16;   T = 32;
  CRATE_N_PROCX = 4;  CRATE_N_PROCY = 8; CRATE_N_PROCZ = 8;

  // Read arguments from command line

  while (argc > 3 && arg[1][0] == '-')
    {
      if (arg[1][2]) usage();

      switch (arg[1][1])
	{
	case 'h':
	  usage();
	  exit(0);
	case 'L':
	  if (argc < 3) usage();
	  L = atoi(arg[2]);
	  argc -= 2; arg += 2; break;
	case 'T':
	  if (argc < 3) usage();
	  T = atoi(arg[2]);
	  argc -= 2; arg += 2; break;
	case 'X':
	  if (argc < 3) usage();
	  CRATE_N_PROCX = atoi(arg[2]);
	  argc -= 2; arg += 2; break;
	case 'Y':
	  if (argc < 3) usage();
	  CRATE_N_PROCY = atoi(arg[2]);
	  argc -= 2; arg += 2; break;
	case 'Z':
	  if (argc < 3) usage();
	  CRATE_N_PROCZ = atoi(arg[2]);
	  argc -= 2; arg += 2; break;
	case 'b':
	  if (argc < 3) usage();
	  g_beta = atof(arg[2]);
	  argc -= 2; arg += 2; break;
	case 'k':
	  if (argc < 3) usage();
	  g_kappa = atof(arg[2]);
	  argc -= 2; arg += 2; break;
	case 'm':
	  if (argc < 3) usage();
	  g_mu = atof(arg[2]);
	  argc -= 2; arg += 2; break;
	default:
	  usage();
	  exit(1);
	}
    }
  
  // Read in and out file from command line

  switch (argc)
    {
    case 3:
      infile  = arg[1];
      outfile = arg[2];
      break;
    default:
      usage();
      exit(1);
    }
  
  // Initialize lattice and call for memory

  initialize();
  geometry();
  status = init_read_crate_gauge_field();

  // Start reading

  printf("\n Read configuration: \n");
  printf("-------------------------\n");
  printf(" format : apeNEXT crate\n");
  printf("   file : %s\n",infile);
  status = read_crate_gauge_field(infile);
  if (status != 0) die("Error while reading apeNEXT configuration",500);

  // Calculate (average) plaquette and print parameters

  plaq = plaquette();
  printf("   plaq : %e\n",plaq);
  printf("   beta : %f\n",g_beta);
  printf("   kappa: %f\n",g_kappa);
  printf("     mu : %f\n",g_mu);

  // Write configuration in ILDG format

  printf("\n Write configuration: \n");
  printf("-------------------------\n");
  printf(" format : ILDG\n");
  printf("   file : %s\n",outfile);
  status = write_lime_gauge_field(outfile, plaq, 1);
  if (status != 0) die("Error while writing ILDG configuration", 500);


  // Uncomment for testing purpose only (reads ILDG configuration in again)
  
  /*  
   printf("\n (Testing) Read ILDG configuration again: \n"); 
   printf("--------------------------------------------\n");
   plaq = 0;
   status = read_lime_gauge_field(outfile);
   plaq = plaquette();
   printf("   plaq : %e\n",plaq); 
  */
  

  // Release memory
  finalize_read_crate_gauge_field();
  finalize();

  return 0;
}



//============================================================================
