
/*******************************************************************************
 * $Id: geometry.c,v 1.12 2004/01/19 17:46:23 knagai Exp $
 *
 * File geometry.c
 *
 * Subroutines related to the lattice geometry
 *
 * The externally accessible function is
 *
 *   void geometry(void)
 *     Computes the index arrays ipt, iup and idn
 *
 * Version: 1.0
 * Author: Martin Luescher <luscher@mail.desy.ch>
 * Date: 24.10.2000
 *
 *
 *
 *  Check and correction in the parallelisation version (MPI version)
 *    done by Kei-ichi Nagai  
 *
 *******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "su3.h"
#include "global.h"

#if defined MPI
/* rather naive geometry: 
   the points are ordered in the following way
   first we have the points in the inner T*LX*LY*LZ box 
   then the T*LY*LZ slices at x1=-1 and x1=LX
   then the T*LY*LZ slices at x2=-1 and x2=LY
   then the T*LX*LZ slices at x3=-1 and x3=LZ
   finally the corners (this we only need for the gauge update)
   for x1=-1 and x1=LX;  x2=-1 and x2=LY
   for x1=-1 and x1=LX;  x3=-1 and x3=LY
   for x2=-1 and x2=LZ;  x3=-1 and x3=LY
*/

int index(const int x0, const int x1, const int x2, const int x3){
  int y0,y1,y2,y3;
  int ix;
  /* periodic boundary conditions: */
  y0 = (x0 + T ) % T; 
  y1 = (x1 + LX) % LX; 
  y2 = (x2 + LY) % LY; 
  y3 = (x3 + LZ) % LZ;
  ix = ((y0*LX + y1)*LY + y2)*LZ + y3;

  /*take care of the parallelization*/
#if ((defined PARALLEL1) || (defined PARALLEL2) || (defined PARALLEL3))
  if(x1 == -1){
    ix = VOLUME+y0*LY*LZ+y2*LZ+y3;
  }
  if(x1 == LX){
    ix = VOLUME+T*LY*LZ+y0*LY*LZ+y2*LZ+y3;
  }

  /* unnecessary ?? */
  if(x1==-2){
      y1=(x1+LX)%LX ;
      y1+=1 ;
      ix = ((y0*LX + y1)*LY + y2)*LZ + y3;
  }
  if(x1==LX+1){
      y1=(x1+LX)%LX ;
      y1-=1 ;
      ix = ((y0*LX + y1)*LY + y2)*LZ + y3;
  }

#endif
#if ((defined PARALLEL2)||(defined PARALLEL3))
  if(x2 == -1){
    ix = VOLUME+2*T*LY*LZ+y0*LX*LZ+y1*LZ+y3;
  }
  if(x2 == LY){
    ix = VOLUME+2*T*LY*LZ+T*LX*LZ+y0*LX*LZ+y1*LZ+y3;
  }
#endif
#if defined PARALLEL3
  if(x3 == -1){
    ix = VOLUME+2*T*LY*LZ+2*T*LX*LZ+y0*LX*LY+y1*LY+y2;
  }
  if(x3 == LZ){
    ix = VOLUME+2*T*LY*LZ+2*T*LX*LZ+T*LX*LY+y0*LX*LY+y1*LY+y2;
  }
#endif
  /* take care of the corners; only needed for the 
     parallelization of the gauge-update */
#if ((defined PARALLEL2)||(defined PARALLEL3))
  if(x1 == -1){
    if(x2 == -1){
      ix = VOLUME+RAND+y0*LZ+y3;
    }
    if(x2 == LY){
      ix = VOLUME+RAND+T*LZ+y0*LZ+y3;
    }
  }
  if(x1 == LX){
    if(x2 == -1){
      ix = VOLUME+RAND+2*T*LZ+y0*LZ+y3;
    }
    if(x2 == LY){
      ix = VOLUME+RAND+3*T*LZ+y0*LZ+y3;
    }
  }
#endif 
#if defined PARALLEL3
  if(x1 == -1){
    if(x3 == -1){
      ix = VOLUME+RAND+4*T*LZ+y0*LY+y2;
    }
    if(x3 == LZ){
      ix = VOLUME+RAND+4*T*LZ+T*LY+y0*LY+y2;
    }
  }
  if(x1 == LX){
    if(x3 == -1){
      ix = VOLUME+RAND+4*T*LZ+2*T*LY+y0*LY+y2;
    }
    if(x3 == LZ){
      ix = VOLUME+RAND+4*T*LZ+3*T*LY+y0*LY+y2;
    }
  }
  if(x2 == -1){
    if(x3 == -1){
      ix = VOLUME+RAND+4*T*LZ+4*T*LY+y0*LX+y1;
    }
    if(x3 == LZ){
      ix = VOLUME+RAND+4*T*LZ+4*T*LY+T*LX+y0*LX+y1;
    }
  }
  if(x2==LY){
    if(x3 == -1){
      ix = VOLUME+RAND+4*T*LZ+4*T*LY+2*T*LX+y0*LX+y1;
    }
    if(x3 == LZ){
      ix = VOLUME+RAND+4*T*LZ+4*T*LY+3*T*LX+y0*LX+y1;
    }
  }
#endif 

  return ix;
}

void geometry(void){
  int t,x,y,z,ix;
  int ry,rz;

  ix=0;
  for (t = 0; t < T; t++){
    for (x = 0; x < LX; x++){
      for (y = 0; y < LY; y++){
        for (z = 0; z < LZ; z++){
          g_ipt[t][x][y][z]=ix;
	  ix++;
	}
      }
    }
  }

  ry = 0; 
  rz = 0;
#if ((defined PARALLEL2)||(defined PARALLEL3))
  ry = 1;
#endif
#if defined PARALLEL3
  rz = 1;
#endif

  for (t = 0; t < T; t++){
    for (x=-1;x<LX+1;x++){
      for (y=-ry;y<LY+ry;y++){
        for (z=-rz;z<LZ+rz;z++){
          ix=index(t,x,y,z);
	  
          g_iup[ix][0]=index(t+1,x,y,z);
          g_idn[ix][0]=index(t-1,x,y,z);
	  
          g_iup[ix][1]=index(t,x+1,y,z);
          g_idn[ix][1]=index(t,x-1,y,z);
  
          g_iup[ix][2]=index(t,x,y+1,z);
          g_idn[ix][2]=index(t,x,y-1,z);
	  
          g_iup[ix][3]=index(t,x,y,z+1);
          g_idn[ix][3]=index(t,x,y,z-1);

        }
      }
    }
  }
}

/* the following else belong to ifdef MPI*/
#else

static int init=0,nl=1;

static int safe_mod(const int x, const int y){
  if (y<=0){
    printf("Error in subroutine safe_mod: improper arguments\n");
    printf("Program aborted\n");
    exit(0);
  }

  if (x>=0){
    return(x%y);
  }
  else{
    return((y-(abs(x)%y))%y);
  }
}


int index(const int x0, const int x1, const int x2, const int x3){
  int y0,y1,y2,y3;
  int xb1,xb2,xb3;
  int xn1,xn2,xn3;
  int ib,in, bl;

  bl = L;
  
  if (init == 0){
    if (L%4 == 0){
      bl = 4;
    }
    else if (L%3 == 0){
      bl = 3;
    }
    else if (L%2 == 0){
      bl = 2;
    }
    
    nl = L/bl; 
    init = 1;
  }
  
  y0 = safe_mod(x0,T);
  y1 = safe_mod(x1,L);
  y2 = safe_mod(x2,L);
  y3 = safe_mod(x3,L);
  
  xb1 = y1%bl;
  xb2 = y2%bl;
  xb3 = y3%bl;
  
  xn1 = y1/bl;
  xn2 = y2/bl;
  xn3 = y3/bl;   
  
  ib = xb3+xb2*bl+xb1*bl*bl+y0*bl*bl*bl;
  in = xn1+xn2*nl+xn3*nl*nl;
  
  return(ib+in*T*bl*bl*bl);
}


void geometry(void){
  int x0,x1,x2,x3,ix;
  
  for (x0 = 0; x0 < T; x0++){
    for (x1 = 0; x1 < L; x1++){
      for (x2 = 0; x2 < L; x2++){
	for (x3 = 0; x3 < L; x3++){
	  ix = index(x0,x1,x2,x3);
	  g_ipt[x0][x1][x2][x3] = ix;
	  
	  g_iup[ix][0]=index(x0+1,x1,x2,x3);
	  g_idn[ix][0]=index(x0-1,x1,x2,x3);
	  
	  g_iup[ix][1]=index(x0,x1+1,x2,x3);
	  g_idn[ix][1]=index(x0,x1-1,x2,x3);
	  
	  g_iup[ix][2]=index(x0,x1,x2+1,x3);
	  g_idn[ix][2]=index(x0,x1,x2-1,x3);
	  
	  g_iup[ix][3]=index(x0,x1,x2,x3+1);
	  g_idn[ix][3]=index(x0,x1,x2,x3-1);
	}
      }
    }
  }
}
/* the following endif belong to ifdef MPI*/
#endif
