/***************************************************************************
! io.c
!
! collection of routines for 
!  + reading apeNEXT-crate configuration (based on "crate2gwc" by Ines Wetzorke)
!  + writing/reading ILDG configuration  (based on "io.c"     by Carsten Urbach)
!
! author: Andre Sternbeck <andre.sternbeck@physik.hu-berlin.de
****************************************************************************/

#define _LARGEFILE_SOURCE
#define _FILE_OFFSET_BITS 64

#include<stdlib.h>
#include<stdio.h>
#include<malloc.h>
#include<string.h>
#ifdef MPI
#include<unistd.h>
#endif
#include "lime.h"
#include "su3.h"
#include "global.h"
#include "io.h"


#define MAXBUF 1048576

void  byte_swap_assign(void * out_ptr, void * in_ptr, int nmemb);
void  float2double(double *dest, float *source, int n);
void  double2double(double *dest, double *source, int n);
int   write_ildg_format_xml(char *filename, LimeWriter *limewriter, double plaq);
off_t file_size(FILE *fp);

int     *e_to_lex, *o_to_lex;
double  *u;
double  *v;
double  *w;

#define LNODE_X (L/CRATE_N_PROCX)
#define LNODE_Y (L/CRATE_N_PROCY)
#define LNODE_Z (L/CRATE_N_PROCZ)
#define NT      T

#define SIZE       LNODE_X*LNODE_Y*LNODE_Z
#define SIZEH      SIZE/2

#define UNIT       DIM*2*NCOL*NCOL
#define VOL        LNODE_X*LNODE_Y*LNODE_Z*NT

static int init_crate_gauge_field = 0;

int init_read_crate_gauge_field(void)
{
  int   neven, nodd, site, z,y,x,t, sum, rest; 

  if (init_crate_gauge_field != 0) return;

  u = (double *) malloc ( SIZE*UNIT*sizeof(double) );
  if ( !u ) {
    fprintf ( stderr, "Not enough memory for u.\n" );
    exit(2);
  }

  v = (double *) malloc ( VOL*UNIT*sizeof(double) );
  if ( !v ) {
    fprintf ( stderr, "Not enough memory for v.\n" );
    exit(2);
  }

  w = (double *) malloc (VOLUME*UNIT*sizeof(double) );
  if ( !w ) {
    fprintf ( stderr, "Not enough memory for w.\n" );
    exit(2);
  }

  e_to_lex = (int *) malloc (SIZEH*sizeof(int) );
  o_to_lex = (int *) malloc (SIZEH*sizeof(int) );

  neven = 0;
  nodd  = 0;
  site  = 0;
  
  for ( z=0 ; z<LNODE_Z ; z++ ) {
    for ( y=0 ; y<LNODE_Y ; y++ ) {
      for ( x=0 ; x<LNODE_X ; x++ ) {
	
	sum = x+y+z;
	rest = sum%2;
	
	if ( rest == 0 ) e_to_lex [neven++] = site;
	else             o_to_lex [nodd++ ] = site;
	
	site++;
      }
    }
  }
  init_crate_gauge_field = 1;
  return 0;
}

int read_crate_gauge_field(char *filename)
{
  int     site, z,y,x,t, sum, rest, status;
  int     kx, ky, kz, gt, gz, gy, gx, i, ix, mu;
  FILE    *ifs;
  su3     *tmp;

double tr;
  ifs = fopen(filename, "rb" );
  status = init_read_crate_gauge_field();


  for ( kx=0 ; kx<CRATE_N_PROCX ; kx++ ) {
    for ( ky=0 ; ky<CRATE_N_PROCY ; ky++ ) {
      for ( kz=0 ; kz<CRATE_N_PROCZ ; kz++ ) {
	
	ix=0;
	for ( gt=0 ; gt<NT ; gt++ ) {

          fread ( u, UNIT*sizeof(double), SIZE, ifs );

	  for ( i=0 ; i<SIZEH ; i++ ) {
	    double2double( &v[ (e_to_lex[i]+gt*SIZE)*UNIT ], &u[ i*UNIT ],UNIT );
          }

          for ( i=0 ; i<SIZEH ; i++ ) {
	    double2double( &v[ (o_to_lex[i]+gt*SIZE)*UNIT ], &u[ (i+SIZEH)*UNIT ],UNIT );
          }

	  for ( gz=0 ; gz<LNODE_Z ; gz++ ) {
	    for ( gy=0 ; gy<LNODE_Y ; gy++ ) {
	      for ( gx=0 ; gx<LNODE_X ; gx++ ) {
		
		/* global coordinates */
		x = kx*LNODE_X + gx;
		y = ky*LNODE_Y + gy;
		z = kz*LNODE_Z + gz;
		t =         gt;
		
		/* new index */
		site = x * T * LZ * LY
                     + y * T * LZ
                     + z * T
                     + t;

#ifdef LITTLE_ENDIAN
		for (mu=0; mu<DIM; mu++)
		  {
		    tmp = (su3*)&v[ix*UNIT+mu*2*9];
	    	    _su3_assign(g_gauge_field[g_ipt[t][x][y][z]][mu],*tmp);
		  }
#else
	???	TO BE IMPLEMENTED YET
	???	byte_swap_assign( &w[site*UNIT], &v[ix*UNIT],UNIT );
#endif
		ix++;

	      }
	    }
	  }
	}
	/* extra time slice in TAO code */
	fread ( u, UNIT*sizeof(double), SIZE, ifs );
      }
    }
  }
  fclose ( ifs );
}

int finalize_read_crate_gauge_field(void)
{
  free(e_to_lex);
  free(o_to_lex);
  free(w); free(u); free(v);
}

int write_lime_gauge_field(char *filename, const double plaq, const int counter)
{
  FILE             *ofs;
  LimeWriter       *limewriter;
  LimeRecordHeader *limeheader;

  /* Message end and Message begin flag */
  int   ME_flag=0, MB_flag=0, status=0;
  int   t=0, x=0, y=0, z=0;
  char  message[100];
  off_t bytes;
#ifdef LITTLE_ENDIAN
  su3 tmp[4];
#endif

  sprintf(message,"\n plaquette = %e\n trajectory nr = %d\n beta = %f, kappa = 0.0, mu = %f, c2_rec = 0.0", 
	  plaq, counter, g_beta, g_mu);
  bytes = strlen( message );
  ofs = fopen(filename, "w");

  if(ofs == (FILE*)NULL) {
    fprintf(stderr, "Could not open file %s for writing!\n Aborting...\n", 
            filename);
    exit(500);
  }
  limewriter = limeCreateWriter( ofs );
  if(limewriter == (LimeWriter*)NULL) {
    fprintf(stderr, "LIME error in file %s for writing!\n Aborting...\n", 
	    filename);
    exit(500);
  }
  write_ildg_format_xml("temp.xml", limewriter, plaq);

  limeheader = limeCreateHeader(MB_flag, ME_flag, "xlf-info", bytes);
  status     = limeWriteRecordHeader( limeheader, limewriter);
  if(status < 0 ) {
    fprintf(stderr, "LIME write header error %d\n", status);
    exit(500);
  }
  limeDestroyHeader( limeheader );
  limeWriteRecordData(message, &bytes, limewriter);

  bytes   = LX*LY*LZ*T*4*sizeof(su3);
  MB_flag = 0; ME_flag = 1;
  limeheader = limeCreateHeader(MB_flag, ME_flag, "ildg-binary-data", bytes);
  status     = limeWriteRecordHeader( limeheader, limewriter);
  if(status < 0 ) {
    fprintf(stderr, "LIME write header error %d\n", status);
    exit(500);
  }
  limeDestroyHeader( limeheader );

  bytes = sizeof(su3);
  for(t = 0; t < T; t++){
    for(z = 0; z < LZ; z++){
      for(y = 0; y < LY; y++){
	for(x = 0; x < LX; x++){
#ifdef LITTLE_ENDIAN
	  byte_swap_assign(tmp, g_gauge_field[ g_ipt[t][x][y][z] ], 
			   4*sizeof(su3)/8); 
	  status = limeWriteRecordData((void*)&tmp[1], &bytes, limewriter);
	  status = limeWriteRecordData((void*)&tmp[2], &bytes, limewriter);
	  status = limeWriteRecordData((void*)&tmp[3], &bytes, limewriter);
	  status = limeWriteRecordData((void*)&tmp[0], &bytes, limewriter);
#else
	  status = limeWriteRecordData((void*)&g_gauge_field[ g_ipt[t][x][y][z] ][1], &bytes, limewriter);
	  status = limeWriteRecordData((void*)&g_gauge_field[ g_ipt[t][x][y][z] ][2], &bytes, limewriter);
	  status = limeWriteRecordData((void*)&g_gauge_field[ g_ipt[t][x][y][z] ][3], &bytes, limewriter);
	  status = limeWriteRecordData((void*)&g_gauge_field[ g_ipt[t][x][y][z] ][0], &bytes, limewriter);
#endif
	  if(status < 0 ) {
	    fprintf(stderr, "LIME write error %d\n", status);
	    exit(500);
	  }
	}
      }
    }
  }
  limeDestroyWriter( limewriter );
  fclose(ofs);
  return(0);
}


int read_lime_gauge_field(char * filename)
{
  FILE   *ifs;
  int    t, x, y, z, status;
  off_t  bytes;
  char   *header_type;
  LimeReader * limereader;
#ifdef LITTLE_ENDIAN
  su3 tmp[4];
#endif

  ifs = fopen(filename, "r");
  if(ifs == (FILE *)NULL) {
    fprintf(stderr, "Could not open file %s\n Aborting...\n", filename);
#ifdef MPI
    MPI_Abort(MPI_COMM_WORLD, 1);
    MPI_Finalize();
#endif
    exit(500);
  }
  limereader = limeCreateReader( ifs );
  if( limereader == (LimeReader *)NULL ) {
    fprintf(stderr, "Unable to open LimeReader\n");
#ifdef MPI
    MPI_Abort(MPI_COMM_WORLD, 1);
    MPI_Finalize();
#endif
    exit(500);
  }
  while( (status = limeReaderNextRecord(limereader)) != LIME_EOF ) {
    if(status != LIME_SUCCESS ) {
      fprintf(stderr, "limeReaderNextRecord returned error with status = %d!\n",
	      status);
      fprintf(stderr, "trying old deprecated file format!\n");
      status = LIME_EOF;
      break;
    }
    header_type = limeReaderType(limereader);
    if(!strcmp("ildg-binary-data",header_type)) break;
  }
  if(status == LIME_EOF) {
    fprintf(stderr, "no ildg-binary-data record found in file %s\n",filename);
    //   fprintf(stderr, "trying old deprecated file format!\n");
    limeDestroyReader(limereader);
    fclose(ifs);
    //  read_gauge_field(filename);
    return(0);
  }
  bytes = limeReaderBytes(limereader);
  if((int)bytes != LX*N_PROC_X*LY*N_PROC_Y*LZ*N_PROC_Z*T*4*sizeof(su3)) {
    fprintf(stderr, "Probably wrong lattice size or precision in file %s\n",
	    filename);
    fprintf(stderr, "Aborting...!\n");
#ifdef MPI
    MPI_Abort(MPI_COMM_WORLD, 1);
    MPI_Finalize();
#endif
    exit(501);
  }

  bytes = 4*sizeof(su3);
#ifndef LITTLE_ENDIAN
  bytes = sizeof(su3);
#endif
  for(t = 0; t < T; t++){
    for(z = 0; z < LZ; z++){
      for(y = 0; y < LY; y++){
#if (defined MPI && (defined PARALLEL1 || defined PARALLEL2 || defined PARALLEL3))
	limeReaderSeek(limereader, (off_t)
		       (g_proc_coords[0]*LX + 
			((t*LZ*N_PROC_Z+g_proc_coords[2]*LZ+z)*LY*N_PROC_Y 
			 + g_proc_coords[1]*LY+y)*LX*N_PROC_X)*4*sizeof(su3),
		       SEEK_SET);
#endif
	for(x = 0; x < LX; x++){
#ifdef LITTLE_ENDIAN
	  status = limeReaderReadData(tmp, &bytes, limereader);
	  byte_swap_assign(&g_gauge_field[ g_ipt[t][x][y][z] ][0], &tmp[3], 
			   sizeof(su3)/8);
	  byte_swap_assign(&g_gauge_field[ g_ipt[t][x][y][z] ][1], &tmp[0], 
			   sizeof(su3)/8);
	  byte_swap_assign(&g_gauge_field[ g_ipt[t][x][y][z] ][2], &tmp[1], 
			   sizeof(su3)/8);
	  byte_swap_assign(&g_gauge_field[ g_ipt[t][x][y][z] ][3], &tmp[2], 
			   sizeof(su3)/8);
#else
	  status = limeReaderReadData(&g_gauge_field[ g_ipt[t][x][y][z] ][1], 
				      &bytes, limereader);
	  status = limeReaderReadData(&g_gauge_field[ g_ipt[t][x][y][z] ][2], 
				      &bytes, limereader);
	  status = limeReaderReadData(&g_gauge_field[ g_ipt[t][x][y][z] ][3], 
				      &bytes, limereader);
	  status = limeReaderReadData(&g_gauge_field[ g_ipt[t][x][y][z] ][0], 
				      &bytes, limereader);
#endif
	  if(status < 0 && status != LIME_EOR) {
	    fprintf(stderr, "LIME read error occured with status = %d while reading file %s!\n Aborting...\n", status, filename);
#ifdef MPI
	    MPI_Abort(MPI_COMM_WORLD, 1);
	    MPI_Finalize();
#endif
	    exit(500);
	  }
	}
      }
    }
  }
  limeDestroyReader(limereader);
  fclose(ifs);
  return(0);
}

int write_ildg_format_xml(char *filename, LimeWriter * limewriter, double plaq)
{
  FILE * ofs;
  off_t bytes, bytes_left, bytes_to_copy;
  int MB_flag=1, ME_flag=0, status=0;
  LimeRecordHeader * limeheader;
  char buf[MAXBUF];

  ofs = fopen(filename, "w");

  fprintf(ofs, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  fprintf(ofs, "<ildgFormat xmlns=\"http://www.lqcd.org/ildg\"\n");
  fprintf(ofs, "            xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n");
  fprintf(ofs, "            xsi:schemaLocation=\"http://www.lqcd.org/ildg filefmt.xsd\">\n");
  fprintf(ofs, "  <version> 1.0 </version>\n");
  fprintf(ofs, "  <field> su3gauge </field>\n");
  fprintf(ofs, "  <precision> 64 </precision>\n");
  fprintf(ofs, "  <avePlaquette> %e </avePlaquette>\n", plaq);
  fprintf(ofs, "  <lx> %d </lx>\n", LX*N_PROC_X);
  fprintf(ofs, "  <ly> %d </ly>\n", LY*N_PROC_Y);
  fprintf(ofs, "  <lz> %d </lz>\n", LZ*N_PROC_Z);
  fprintf(ofs, "  <lt> %d </lt>\n", T);
  fprintf(ofs, "</ildgFormat>");
  fclose( ofs );
  ofs = fopen(filename, "r");
  bytes = file_size(ofs);
  
  limeheader = limeCreateHeader(MB_flag, ME_flag, "ildg-format", bytes);
  if(limeheader == (LimeRecordHeader*)NULL) {
    fprintf(stderr, "LIME create header ildg-format error\n Aborting...\n");
    exit(500);
  }
  status = limeWriteRecordHeader( limeheader, limewriter);
  if(status < 0 ) {
    fprintf(stderr, "LIME write header ildg-format error %d\n Aborting...\n", 
	    status);
    exit(500);
  }
  limeDestroyHeader( limeheader );

  /* Buffered copy */
  bytes_left = bytes;
  while(bytes_left > 0){
    if(MAXBUF < bytes_left) {
      bytes_to_copy = MAXBUF;
    }
    else bytes_to_copy = bytes_left;
    if( bytes_to_copy != fread(buf,1,bytes_to_copy,ofs))
      {
	fprintf(stderr, "Error reading %s\n", filename);
	return EXIT_FAILURE;
      }
    
    status = limeWriteRecordData(buf, &bytes_to_copy, limewriter);
    if (status != 0) {
      fprintf(stderr, "LIME error writing ildg-format status = %d\n Aborting...\n", status);
      exit(500);
    }
    bytes_left -= bytes_to_copy;
  }
  

  fclose( ofs );
  return(0);
}

off_t file_size(FILE *fp)
{
  off_t oldpos = ftello(fp);
  off_t length;
  
  if (fseeko(fp, 0L,SEEK_END) == -1)
    return -1;
  
  length = ftello(fp);
  
  return ( fseeko(fp,oldpos,SEEK_SET) == -1 ) ? -1 : length;
  
}

void byte_swap_assign(void * out_ptr, void * in_ptr, int nmemb){
  int j;
  char * char_in_ptr, * char_out_ptr;
  double * double_in_ptr, * double_out_ptr;

  double_in_ptr = (double *) in_ptr;
  double_out_ptr = (double *) out_ptr;
  for(j = 0; j < nmemb; j++){
    char_in_ptr = (char *) double_in_ptr;
    char_out_ptr = (char *) double_out_ptr;
    
    char_out_ptr[7] = char_in_ptr[0];
    char_out_ptr[6] = char_in_ptr[1];
    char_out_ptr[5] = char_in_ptr[2];
    char_out_ptr[4] = char_in_ptr[3];
    char_out_ptr[3] = char_in_ptr[4];
    char_out_ptr[2] = char_in_ptr[5];
    char_out_ptr[1] = char_in_ptr[6];
    char_out_ptr[0] = char_in_ptr[7];
    double_in_ptr++;
    double_out_ptr++;
  }
}

void float2double(double *dest, float *source, int n)
{
 // Copy float to double

  register int i;
  for ( i=0 ; i<n; i++ ) dest[i] = (double) source[i];
}
void double2double(double *dest, double *source, int n)
{
 // Copy float to double

  register int i;
  for ( i=0 ; i<n; i++ ) dest[i] = source[i];
}


//==============================================================================
